"""[summary]

implementing multi-scale DNN MscaleDNN version 2
e.g. 2007.11207, 2009.12729

But with one important difference, the scale is only applied
to the time dimension, other dimensions are left alone.

see https://gitlab.com/SpaceTimeKhantinuum/ml/-/blob/master/waveforms/july2021/time_pars_subnetwork_test/fit_multiple.ipynb
"""

import tensorflow as tf
import numpy as np

# import tensorflow_addons as tfa


def build_subnetwork(
    input_tensor, units, activation="relu", n_blocks=1, layers_per_block=3
):
    """
    subnetwork with skip-connections

    n_blocks >= 1
    """
    assert n_blocks >= 1, f"n_blocks must be >1, got {n_blocks}"

    # tmp = input_tensor
    tmp = tf.keras.layers.Dense(units, activation=activation)(input_tensor)

    x = tf.keras.layers.Dense(units, activation=activation)(input_tensor)
    # x = tfa.layers.GroupNormalization(groups=1)(x)
    # x = tf.keras.layers.BatchNormalization(momentum=0.9)(x)
    for _ in range(n_blocks):
        for _ in range(layers_per_block):
            x = tf.keras.layers.Dense(units, activation=activation)(x)
        # x = tfa.layers.GroupNormalization(groups=1)(x)
        # x = tf.keras.layers.BatchNormalization(momentum=0.9)(x)

        x = tf.keras.layers.add([x, tmp])
        tmp = x

    # single output here?
    x = tf.keras.layers.Dense(1, activation="linear")(x)

    return x


def make_scale_tensor(input_shape, scale_dimension, scale, dtype=None):
    """[summary]

    Args:
        input_shape ([type]): [description]
        scale_dimension ([type]): [description]
        scale ([type]): [description]
        dtype ([type], optional): [description]. Defaults to None.

    Returns:
        [type]: [description]
    """
    scales = np.ones(shape=(input_shape,), dtype=dtype)
    scales[scale_dimension] = scale
    return scales


class ScaleLayer(tf.keras.layers.Layer):
    def __init__(self, scale_tensor, **kwargs):
        super(ScaleLayer, self).__init__()
        self.scale_tensor = scale_tensor

    def call(self, inputs):
        return tf.math.multiply(inputs, self.scale_tensor)

    def get_config(self):
        config = super(ScaleLayer, self).get_config()
        config.update({"scale_tensor": self.scale_tensor})
        return config


# this version supports variable number of units and blocks for each scale.
# the idea is that maybe smaller scales can cope with smaller networks
def build_model(
    input_shape=1,
    output_shape=1,
    units=[128],
    activation="relu",
    n_blocks=[2],
    scales=[1],
    layers_per_block=3,
    scale_dimension=0,
    dtype=None,
):
    """[summary]

    Args:
        input_shape (int, optional): [description]. Defaults to 1.
        output_shape (int, optional): [description]. Defaults to 1.
        units (list, optional): [description]. Defaults to [128].
        activation (str, optional): [description]. Defaults to "relu".
        n_blocks (list, optional): [description]. Defaults to [2].
        scales (list, optional): [description]. Defaults to [1].
        layers_per_block (int, optional): [description]. Defaults to 3.
        scale_dimension (int, optional): [description]. Defaults to 0.
            Use this to select which feature to apply the scaling to.
        dtype ([type], optional): [description]. Defaults to None.

    Returns:
        [type]: [description]
    """
    assert (
        len(scales) == len(n_blocks) == len(units)
    ), "units, n_blocks, scales must have same length"

    input_layer = tf.keras.Input(shape=(input_shape,))

    # create sub-networks
    xs = []
    for i, scale in enumerate(scales):
        scale_tensor = make_scale_tensor(
            input_shape, scale_dimension, scale, dtype=dtype
        )
        scaled_input = ScaleLayer(scale_tensor)(input_layer)
        xs.append(
            build_subnetwork(
                input_tensor=scaled_input,
                units=units[i],
                activation=activation,
                n_blocks=n_blocks[i],
                layers_per_block=layers_per_block,
            )
        )

    if len(xs) > 1:
        output_layer = tf.keras.layers.add([x for x in xs])
    else:
        output_layer = xs[0]

    output_layer = tf.keras.layers.Dense(output_shape, activation="linear")(
        output_layer
    )
    model = tf.keras.models.Model(inputs=input_layer, outputs=output_layer)
    return model


# based on the work by Dongwei Chen+ 2020
# "Deep Residual Learning for Nonlinear Regression"
# https://www.mdpi.com/1099-4300/22/2/193
#  https://github.com/DowellChan/ResNetRegression

# https://github.com/tiantiy/ResNetRegression/blob/master/ResNetOptimalModel.py

# modified to use GroupNormalization instead of BatchNormalization
# with groups=1 which is the same as LayerNormalization

from tensorflow.keras import layers
from tensorflow.keras import models
import tensorflow_addons as tfa


def identity_block(
    input_tensor, units, group_norm=True, groups=1, batch_norm=False, momentum=0.9
):
    """The identity block is the block that has no conv layer at shortcut.
    By default uses GroupNormalization after the Dense layer but before
    the ReLU.

    # Arguments
            input_tensor: input tensor
            units:output shape
            group_norm: bool, default True - use GroupNormalization
            groups: int, default 1 - number of groups, defaults to LayerNormalization
            batch_norm: bool, default False - use batch normalization or not
            momentum: float, default 0.9 - BatchNorm momentum parameter
    # Returns
            Output tensor for the block.
    """

    x = layers.Dense(units)(input_tensor)
    if group_norm:
        x = tfa.layers.GroupNormalization(groups=groups)(x)
    if batch_norm:
        x = layers.BatchNormalization(momentum=momentum)(x)
    x = layers.Activation("relu")(x)

    x = layers.Dense(units)(x)
    if group_norm:
        x = tfa.layers.GroupNormalization(groups=groups)(x)
    if batch_norm:
        x = layers.BatchNormalization(momentum=momentum)(x)
    x = layers.Activation("relu")(x)

    x = layers.Dense(units)(x)
    if group_norm:
        x = tfa.layers.GroupNormalization(groups=groups)(x)
    if batch_norm:
        x = layers.BatchNormalization(momentum=momentum)(x)

    x = layers.add([x, input_tensor])
    x = layers.Activation("relu")(x)

    return x


def dens_block(
    input_tensor, units, group_norm=True, groups=1, batch_norm=False, momentum=0.9
):
    """A block that has a dense layer at shortcut.
    # Arguments
            input_tensor: input tensor
            units:output shape
            group_norm: bool, default True - use GroupNormalization
            groups: int, default 1 - number of groups, defaults to LayerNormalization
            batch_norm: bool, default False - use batch normalization or not
            momentum: float, default 0.9 - BatchNorm momentum parameter
    # Returns
            Output tensor for the block.
    """
    x = layers.Dense(units)(input_tensor)
    if group_norm:
        x = tfa.layers.GroupNormalization(groups=groups)(x)
    if batch_norm:
        x = layers.BatchNormalization(momentum=momentum)(x)
    x = layers.Activation("relu")(x)

    x = layers.Dense(units)(x)
    if group_norm:
        x = tfa.layers.GroupNormalization(groups=groups)(x)
    if batch_norm:
        x = layers.BatchNormalization(momentum=momentum)(x)
    x = layers.Activation("relu")(x)

    x = layers.Dense(units)(x)
    if group_norm:
        x = tfa.layers.GroupNormalization(groups=groups)(x)
    if batch_norm:
        x = layers.BatchNormalization(momentum=momentum)(x)

    shortcut = layers.Dense(units)(input_tensor)
    if group_norm:
        shortcut = tfa.layers.GroupNormalization(groups=groups)(shortcut)
    if batch_norm:
        shortcut = layers.BatchNormalization(momentum=momentum)(shortcut)

    x = layers.add([x, shortcut])
    x = layers.Activation("relu")(x)
    return x


def ResNet(
    input_shape,
    output_shape,
    width=64,
    num_blocks=3,
    group_norm=True,
    groups=1,
    batch_norm=False,
    momentum=0.9,
):
    """builds a ResNet regression model

    By default it uses GroupNormalization with groups=1 (i.e. LayerNormalization)
    as the normalization layers.

    # Arguments
            input_shape: number of inputs
            output_shape: number of outputs
            width [64]:
            num_blocks [3]: total number of blocks. A block consists of
                            a dense block followed by two identity blocks.
            group_norm: bool, default True - use GroupNormalization
            groups: int, default 1 - number of groups, defaults to LayerNormalization
            batch_norm: bool, default False - use batch normalization or not
            momentum: float, default 0.9 - BatchNorm momentum parameter
    # Returns
            A Keras model instance.
    """
    assert num_blocks >= 1, f"num_blocks={num_blocks} not allowed. must be >= 1"

    Res_input = layers.Input(shape=(input_shape,))

    x = dens_block(Res_input, width, group_norm=group_norm, groups=groups, batch_norm=batch_norm, momentum=momentum)
    x = identity_block(x, width, group_norm=group_norm, groups=groups, batch_norm=batch_norm, momentum=momentum)
    x = identity_block(x, width, group_norm=group_norm, groups=groups, batch_norm=batch_norm, momentum=momentum)

    if num_blocks > 1:
        for i in range(1, num_blocks):
            x = dens_block(x, width, group_norm=group_norm, groups=groups, batch_norm=batch_norm, momentum=momentum)
            x = identity_block(x, width, group_norm=group_norm, groups=groups, batch_norm=batch_norm, momentum=momentum)
            x = identity_block(x, width, group_norm=group_norm, groups=groups, batch_norm=batch_norm, momentum=momentum)

    if group_norm:
        x = tfa.layers.GroupNormalization(groups=groups)(x)
    if batch_norm:
        x = layers.BatchNormalization(momentum=momentum)(x)
    x = layers.Dense(output_shape, activation="linear")(x)
    model = models.Model(inputs=Res_input, outputs=x)

    return model



"""[summary]

implementing multi-scale DNN MscaleDNN version 2

e.g. 2007.11207, 2009.12729

see https://gitlab.com/SpaceTimeKhantinuum/ml/-/blob/master/waveforms/april2021/mscale-tests/one-waveform.ipynb
"""

import tensorflow as tf

# import tensorflow_addons as tfa


def build_subnetwork(
    input_tensor, units, activation="relu", n_blocks=1, layers_per_block=3
):
    """
    subnetwork with skip-connections

    n_blocks >= 1
    """
    assert n_blocks >= 1, f"n_blocks must be >1, got {n_blocks}"

    # tmp = input_tensor
    tmp = tf.keras.layers.Dense(units, activation=activation)(input_tensor)

    x = tf.keras.layers.Dense(units, activation=activation)(input_tensor)
    # x = tfa.layers.GroupNormalization(groups=1)(x)
    # x = tf.keras.layers.BatchNormalization(momentum=0.9)(x)
    for _ in range(n_blocks):
        for _ in range(layers_per_block):
            x = tf.keras.layers.Dense(units, activation=activation)(x)
        # x = tfa.layers.GroupNormalization(groups=1)(x)
        # x = tf.keras.layers.BatchNormalization(momentum=0.9)(x)

        x = tf.keras.layers.add([x, tmp])
        tmp = x

    # single output here?
    x = tf.keras.layers.Dense(1, activation="linear")(x)

    return x


class ScaleLayer(tf.keras.layers.Layer):
    def __init__(self, scale, **kwargs):
        super(ScaleLayer, self).__init__()
        self.scale = scale

    def call(self, inputs):
        return inputs * self.scale

    def get_config(self):
        config = super(ScaleLayer, self).get_config()
        config.update({"scale": self.scale})
        return config


# this version supports variable number of units and blocks for each scale.
# the idea is that maybe smaller scales can cope with smaller networks
def build_model(
    input_shape=1,
    output_shape=1,
    units=[128],
    activation="relu",
    n_blocks=[2],
    scales=[1],
    layers_per_block=3,
):
    assert (
        len(scales) == len(n_blocks) == len(units)
    ), "units, n_blocks, scales must have same length"

    input_layer = tf.keras.Input(shape=(input_shape,))

    # create sub-networks
    xs = []
    for i, scale in enumerate(scales):
        scaled_input = ScaleLayer(scale)(input_layer)
        xs.append(
            build_subnetwork(
                input_tensor=scaled_input,
                units=units[i],
                activation=activation,
                n_blocks=n_blocks[i],
                layers_per_block=layers_per_block,
            )
        )

    if len(xs) > 1:
        output_layer = tf.keras.layers.add([x for x in xs])
    else:
        output_layer = xs[0]

    output_layer = tf.keras.layers.Dense(output_shape, activation="linear")(
        output_layer
    )
    model = tf.keras.models.Model(inputs=input_layer, outputs=output_layer)
    return model


"""[summary]

This model is quite similar to mscalev3.py
but I've implemented a change to the architecture.

In each of the sub-networks there are blocks of layers.
At the end of each block there is a branched "output" layer.
It is not an actual output of the network though.
Instead it is a residual node, at the end of the block we sum up
all these branched "output" nodes to give the output prediction
from each sub-network.

In addition, each block also has a skip-connection.

Then each output from the sub-networks are themselves summed up to produce
the final output prediction.

The reasoning is that this produces a residual connection that is
easier to understand in the regression sense.

"""

import tensorflow as tf
import tensorflow_addons as tfa
from wispy.mscalev3 import make_scale_tensor, ScaleLayer

# this version supports variable number of units and blocks for each scale.
# the idea is that maybe smaller scales can cope with smaller networks
def build_model(
    input_shape=1,
    output_shape=1,
    units=128,
    activation="relu",
    n_blocks=2,
    scales=[1],
    layers_per_block=3,
    scale_activation="relu",
    scale_dimension=0,
    dtype=None,
):
    """[summary]

    Args:
        input_shape (int, optional): [description]. Defaults to 1.
        output_shape (int, optional): [description]. Defaults to 1.
        units (int, optional): [description]. Defaults to 128.
        activation (str, optional): [description]. Defaults to "relu".
            The activation function to use in the rest of the network
        n_blocks (int, optional): [description]. Defaults to 2.
        scales (list, optional): [description]. Defaults to [1].
        layers_per_block (int, optional): [description]. Defaults to 3.
        scale_activation (str, optional): [description]. Defaults to "relu".
            The activation function to use after the ScaleLayer, Dense Layers
        scale_dimension (int, optional): [description]. Defaults to 0.
            Use this to select which feature to apply the scaling to.
        dtype ([type], optional): [description]. Defaults to None.

    Returns:
        [type]: [description]
    """
    input_layer = tf.keras.Input(shape=(input_shape,))

    outputs = [None] * len(scales)

    for i, scale in enumerate(scales):
        scale_tensor = make_scale_tensor(
            input_shape, scale_dimension, scale, dtype=dtype
        )
        scaled_input = ScaleLayer(scale_tensor)(input_layer)

        x = tf.keras.layers.Dense(units, activation=scale_activation)(scaled_input)
        # x = tf.keras.layers.Activation(scale_activation)(x)

        outputs[i] = []

        for k in range(n_blocks):
            tmp = x
            for j in range(layers_per_block):
                x = tf.keras.layers.Dense(units, activation=activation)(x)
                # x = tf.keras.layers.Activation(activation)(x)
            # x = tfa.layers.GroupNormalization(groups=1)(x)
            x = tf.keras.layers.add([x, tmp])
            outputs[i].append(tf.keras.layers.Dense(output_shape)(x))
        outputs[i] = tf.keras.layers.add([output for output in outputs[i]])

    if len(outputs) > 1:
        outputs = tf.keras.layers.add([output for output in outputs])
    else:
        outputs = outputs[0]

    model = tf.keras.models.Model(inputs=input_layer, outputs=outputs)

    return model